#
# 	Sebastien.Letort@toulouse.inra.fr
#	Created: April 05, 2007
# 	$Id: General.pm 551 2010-08-23 09:55:50Z manu $
#

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

package General;

BEGIN {
	use Exporter   ();
	use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

	$VERSION = do { my @r = (q$Rev: 551 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };

	@ISA         = qw(Exporter);
	@EXPORT      = qw( &TRUE &FALSE Camelize Capitalize GetStreamIn GetStreamOut IsNotNull Max Min MinMax MinMaxArray SafeIsArray SafeIsHash SafeIsa SetIfNull SetUnlessDefined ShowDebug Verbose TryMain ShowStack);
#	%EXPORT_TAGS = ( );     # ex. : TAG => [ qw!name1 name2! ],
#	@EXPORT_OK   = qw( &ShowDebug );
}

use Data::Dumper;
use Scalar::Util;
use Carp;
use IO::File;
use Error qw( :try );

# les constantes
use constant TRUE => 1;
use constant FALSE => 0;

#les fonctions generales

=head2 function Capitalize

  Title        : Capitalize
  Usage        : $string = &Capitalize( $str );
  Prerequisite : none
  Function     : puts all first letters of a string to upper case
                 It works also if words are separated by '_'.
  Returns      : $string, $str with upper case on the first letters of every word
  Args         : $str, a string to capitalize.
  Globals      : none

=cut

sub Capitalize
{
	my ( $str ) = @_;

	my $string = $str;
# 	$string =~ s/([0-9a-zA-Z]+)/\u$1/g;
	$string =~ s/([[:alnum:]]+)/\u$1/g;

	return $string;
}



=head2 function Camelize

  Title        : Camelize
  Usage        : $string = &Camelize( $str );
  Prerequisite : none
  Function     : puts all first letters of a string to upper case,
                 like Capitalize, removes separators '_', ' ' or '''.
                 Other separators are ignored
  Returns      : $string, a one word string in camelcase style.
  Args         : $str, a string to camelize.
  Globals      : none

  TODO         : est-ce qu'on doit enlever tous les separateurs
                 ou bien seulement ceux entre 2 mots ?
                "salut les gars, ca va ?"
                 => 1) "SalutLesGars,CaVa?"
                 => 2) "SalutLesGars, CaVa ?"

                 Pour le moment c'est le 1)

=cut

sub Camelize
{
	my ( $str ) = @_;

	my $string = $str;
# 	$string =~ s/([0-9a-zA-Z]+)/\u$1/g;
	$string =~ s/([[:alnum:]]+)/\u$1/g;
	$string =~ s/[ _']//g;

	return $string;
}

=head2 Procedure TryMain

 Title        : TryMain
 Usage        : 1/ &TryMain();
                2/ &TryMain(1);
                3/ &tryMain(1,\&Main::SomeProcedure());
                4/ &tryMain(\&Main::SomeProcedure(),1);
 Prerequisite : A die, a croak may be thrown
                An error deriving from the Error class may be thrown (see Error.pm package)
 Procedure    : This procedure provides the scripts with a MINIMUM error processing
                Call a function, encapsulating it inside a try/catch construct
				1/ Call the function Main and display minimal message in case of error
				2/ Call the function Main and display the whole stack in case of error
				3/ Same as 2/, but call SomeProcedure instead of Main
				4/ Same as 3/
 Args         : A flag to control the display of the stack (defaults to 0)
				A ref to a function, will be called (defaults to Main::Main)
				The order of the argumens is irrelevant

=cut
sub TryMain
{
	my ($stack_flg,$fn_Main) = (undef, undef);
	
	foreach my $arg (@_)
	{
		if ( ref($arg) eq 'CODE' )
		{
			$fn_Main = $arg;
		}
		if ( ref($arg) eq '' )
		{
			$stack_flg = $arg;
		}
	}
	$stack_flg = 0 unless (defined $stack_flg);
	$fn_Main   = \&main::Main unless (defined $fn_Main);

	try
	{
	    $Error::Debug = 1 if ($stack_flg != 0);
	    &{$fn_Main}();
	    exit 0;
	}
	catch Error with
	{
	     my ( $o_error ) = @_;
	     if ($Error::Debug == 1)
	     {
	         print $o_error->stacktrace();
	     }
	     else
	     {
 			 print "ERROR - FILE  = " . $o_error -> file . "\n";
 			 print "        LINE  = " . $o_error -> line . "\n";
			 print "        VALUE = " . $o_error -> value . "\n";
			 print "        TEXT  = " . $o_error->text . "\n";
			 exit 1;
		 }
	};
}

=head2 Procedure ShowStack

 Title        : ShowStack
 Usage        : &ShowStack();
 Prerequisite : try/catch construct being used, probably through TryMain
 Procedure    : Change the Error::Debug global, so that the stack will be displayed in case of error
                Completely equivalent to passing 1 as $stack_flg in TryMain, exwcept that this function may be
                triggered from Main, thus controlled by some environment variable, parameter, etc.
                See GeneralTestTTryMain.pl for a demo
 Args         : None

=cut
sub ShowStack
{
	$Error::Debug = 1;
}

=head2 Procedure Verbose

 Title        : Verbose
 Usage        : &Verbose( $self, $msg );
                or &Verbose( { verbose => &TRUE }, $msg );
 Prerequisite : none
 Procedure    : warn a message if verbose is true
               	or is $param->{ _verbose } is true.
 Args         : first arg, object with _verbose attribute
                or $hash ref with verbose key
                $msg, string, the message to warn

 Note         : deprecated, try Logger attribute in object context

=cut

sub Verbose
{
	my $param = shift;
	my ( $msg ) = @_;

	my $verbose = &FALSE;
	my $ref = ref $param;
	if( $ref eq 'HASH' )
	{	$verbose = $$param{ verbose };	}
	else
	{	$verbose = $param->{ _verbose };	}

	print STDERR "v-->$msg\n"	if( defined $verbose and &TRUE == $verbose );

	return;
}


=head2 Procedure ShowDebug

	Title      :	ShowDebug
	Usage      :	ShowDebug( deep=>$deep, arg => $ra_args );	<= a faire
	Prerequiste:	none
	Procedure  :	print onto STDERR the name of the function where the function was called
	        	depending on the deep parameter
	        	with the filename and the line number of the current file.
	        	Then it prints the list of arguments with Dumper.
	Args       :	$deep, an integer representing the deep of the call
	        		must be >=1
	        	@a_args, the list of arguments to print out.
	Error      :    none
	Globals    :    none

	Note       :	I used a hash ref to preserved compatibility with previous usage ShowDebug(@a_args)

=cut

sub ShowDebug
{
	my (@args) = @_;

	my $deep = 1;
	if( ref $args[0] eq 'HASH' )
	{
		$rh_hash = shift @args;
		$deep = $$rh_hash{ deep };
	}

	my ($package, $file, $line) = caller();

	my $subroutine = ( caller($deep) )[3] || $package;
	print STDERR "In $subroutine ($file:$line) :\n", Data::Dumper->Dump([@args]);
}


=head2 Function IsNotNull

	Title      :	IsNotNull
	Usage      :	print "\$var n'est pas vide\n"	if( IsNotNull( $var ) );
	Prerequiste:	only works with scalars (strings)
	Function   :	shortcut to know if a variable is defined and non empty
	Returns    :	&TRUE if var is defined and non empty, &FALSE otherwise
	Args       :	$var, a variable that should contains a string.
	Error      :    none
	Globals    :    none

=cut

sub IsNotNull
{
	my ($var) = @_;

	return &TRUE if ( defined $var and $var ne '');
	return &FALSE;
}



=head2 Procedure SetUnlessDefined,SetIfNull

 Title    : SetUnlessDefined,SetIfNull
 Usage    : SetUnlessDefined( \$var, 'some value' );
            SetIfNull( \$var, 'some value' ); 
 Procedure: Set a value to a variable if it hasn't got any.
                     -if not defined for SetUnlessDefined
                     -if not defined or '' for SetIfNull
 Args     : $r_var, a reference to a scalar variable.
            $val, any value
 Error    : none
 Globals  : none

=cut

sub SetUnlessDefined
{
	my ( $r_var, $val ) = @_;

	if( ref( $r_var ) eq 'SCALAR' )
	{
		$$r_var = $val	unless defined ( $$r_var );
		return;
	}
	Carp::croak ( "ERROR - $r_val should be a scalar reference" );
}
sub SetIfNull
{
	my ( $r_var, $val ) = @_;

	if( ref( $r_var ) eq 'SCALAR' )
	{
		$$r_var = $val	unless &IsNotNull ( $$r_var );
		return;
	}
	Carp::croak ( "ERROR - $r_val should be a scalar reference" );
}






=head2 Procedure InitArray

 Title        : InitArray
 Usage        : &InitArray( $ra_array, val=>$val, nb_val => $nb );
 Prerequisite : none
 Procedure    : this function initializes an array
 Args         : $ra_array, a ref to the array to modify
               	$val, the value to affect to elements of the array, default is 0
               	$nb, the number of value to initialize _usefull if your array is still empty
               		default is the length of the array.

 Note         : will move to Array package.

=cut

sub InitArray_OBSOLETE
{
	use Carp;
	carp "obsolete use of General::InitArray, prefer Array::Init\n";

	my $ra_array = shift;
	my %h_args = ( val => 0,
				nb_val => scalar( @$ra_array ),
				@_ );

	my $key;
	for( my $i=0; $i<$h_args{ nb_val }; ++$i )
	{
		$$ra_array[ $i ] = $h_args{ val };
	}

	return;
}






=head2 Procedure SafeIsa

 Usage        : $bool = SafeIsa( $var, $class )
 Function     : check that $var is defined,
                that $var is an object and that $var is a $class instance.
 Args         : $var, any scalar to test,
                $class, string, class name.
 Error        : none.

=cut

sub SafeIsa
{
	my( $var, $class ) = @_;

	return &TRUE if (defined $var &&  &Scalar::Util::blessed( $var ) && $var->isa( $class ) );
	return &FALSE;
} #SafeIsa

=head2 Procedure SafeIsArray,SafeIsHash

 Usage        : $bool = SafeIsArray( $var )
 Function     : check that $var is defined and is a ref to an array/hash
 Args         : $var, any ref-to-array or ref-to-hash to test
 NOTE         : SafeIsScalar does not make sense, because a ref to a scalar whose value is undef is... a ref to scalar
 Error        : none.

=cut

sub SafeIsArray
{
	my ($var) = @_;

	return &TRUE if ( defined $var && ref($var) eq 'ARRAY');
	return &FALSE;
}
sub SafeIsHash
{
	my ($var) = @_;

	return &TRUE if ( defined $var && ref($var) eq 'HASH');
	return &FALSE;
}
 
  
=head2 Procedure MinMaxArray

 Usage        : my ($min,$max) = &MinMaxArray( \@a_data)
 Prerequisite : @a_data is an array or numbers
 Function     : Retrieves the minimum and maximum values of an array, in one pass
 Args         : $ra_data A ref to an array
 Error        : none.

=cut

sub MinMaxArray
{
	my ($ra_data) = @_;
	Carp::croak("ERROR - Input parameter SHOULD BE an array ref") if (! &SafeIsArray( $ra_data ));
	Carp::croak("ERROR - the array is empty") if (@$ra_data == 0);

	my ($min,$max) = ($ra_data->[0], $ra_data->[0]);
	foreach my $data (@{$ra_data})
	{
		$min = $data if ($data < $min);
		$max = $data if ($data > $max);
	}
	return ($min,$max);
}



=head2 Procedure MinMax,Min,Max

 Usage        : my ($min,$max) = &MinMax( $x,$y )
 Prerequisite : $x,$y are two numbers
 Function     : returns ( $min,$max) 
 Args         : $x,$y

=cut
sub MinMax
{
	my ($x,$y) = @_;
	return ($x<$y)?($x,$y):($y,$x);
}
sub Min
{
	my ($x,$y) = @_;
	return ($x<$y)?$x:$y;
}
sub Max
{
	my ($x,$y) = @_;
	return ($x<$y)?$y:$x;
}	
	
=head2 Procedure GetStreamIn,GetStreamOut

 Usage        : my $fh_in = &GetStreamIn( $filepath );
                my $fh_out= &GetStreamOut($filepath );
 Prerequisite : The module File::Which is not required, but recommended
 Function     : Opens the (may be compressed) filepath and returns a file handle, opened in read mode
 Args         : the file path
 Returns      : the file handle
 Global       : %H_COMPRESSION_PRGMS (readonly): key = extension, value = binary name
 
=cut
my %H_COMPRESSION_PRGMS = ( xz => 'xz', lzma => 'lzma', gz => 'gzip', bz2 => 'bzip2', Z => 'compress' );
sub GetStreamIn
{
    my ( $filepath ) = @_;
	return *STDIN if ( $filepath eq '' || $filepath eq '-' || $filepath eq 'stdin');
 
 	my ($ext) = ( $filepath =~ /\.(\w+)$/ );
	my ($exe,$stream);
	$exe = ( defined($ext) && exists($H_COMPRESSION_PRGMS{$ext}) ) ? $H_COMPRESSION_PRGMS{$ext} : '';
	if ($exe ne '')
	{
		my $exe_path = $exe;
		eval ("use File::Which");
	    if ($@ eq "")
	    {
			$exe_path = which($exe);
			if (! defined $exe_path)
			{
				Carp::croak("ERROR - $exe is NOT in the path, or NOT installed");
			}
		}
		$stream = "$exe_path -cd $filepath |";
	}
	else
	{
		$stream = $filepath;
	}
	
	
	my $fh_in = new IO::File($stream) or Carp::croak("ERROR - Can't read >$stream<");
	return $fh_in;
}
sub GetStreamOut
{
    my ( $filepath ) = @_;
    return *STDOUT if ( $filepath eq '' || $filepath eq '-' || $filepath eq 'stdout');
    
	my ($ext) = ($filepath =~ /\.(\w+)$/ );
	my ($exe,$stream);
	$exe = ( defined($ext) && exists($H_COMPRESSION_PRGMS{$ext}) ) ? $H_COMPRESSION_PRGMS{$ext} : '';
	if ($exe ne '')
	{
		eval("use File::Which");
		my $exe_path = $exe;
	    if ($@ eq "")
	    {
			$exe_path = which($exe);
			if (! defined $exe_path)
			{
				Carp::croak("ERROR - $exe is NOT in the path, or NOT installed");
			}
		}
		$stream = "| $exe_path -c >$filepath";
	}
	else
	{
		$stream = ">$filepath";
	}
	
	
	my $fh_out = new IO::File($stream) or Carp::croak("ERROR - Can't write >$stream<");
	return $fh_out;
}	




=head2 Procedure HashWithoutValues_OBSOLETE

 Title        : HashWithoutValues_OBSOLETE
 Usage        : &HashWithoutValues( $rh_hash, @a_values );
 Prerequisite : do not use with string values.
                Works better when there are much more vlaues in the hash than in the list @a_values
 Procedure    : this function will delete keys where values belongs to @a_values
 Args         : $rh_hash, a ref to the hash to modify
                @a_values, a list of values to search

 Note : I suppose that there is another way to do it in one pass
      	like by using grep map or something like that.

=cut

sub HashWithoutValues_OBSOLETE
{
	my ( $rh_hash, @a_values ) = @_;

	foreach my $val ( @a_values )
	{
		foreach my $key ( keys %$rh_hash )
		{
			delete $$rh_hash{ $key }	if( $$rh_hash{ $key } == $val );
		}
	}

	return;
}
1;	# package General
# -------------------------------------------------------------------
# ----------------------- General::Array ----------------------------
# -------------------------------------------------------------------

package Array;

use General;

=head2 function Array::Includes_OBSOLETE

  Title        : Includes_OBSOLETE
  Usage        : @a_indices = &Array::Includes( $ra_list, -values => $ra_values );
               	$indice = &Array::Includes( $ra_list, -val => $val );
  Prerequisite : it use eq to compare values even if it figures.
  Function     : This function checks if the values are into the $ra_list.
  Returns      : @a_indices = a list of indices,
               		the position where the element have been found
               	$indice = the position of $val in $ra_list
               	If one value is not found, its indice will be undef.
  Args         : $ra_list, a ref to the list of reference.
               	%hash, one key
               		-val to seek for one value.
               		-values to seek for many values

=cut

sub Includes_OBSOLETE
{
	my $ra_list = shift;
	my %h_args = (
			-val => undef,
			-values => [],
# 			-cmp => '==',	# no use since eq works for figures
			@_ );

	my $ra_values = $h_args{ -values };
	if( defined( $h_args{ -val } ) )
	{	@$ra_values = ( $h_args{ -val } );	}

	my @a_out = ();
	foreach my $val ( @$ra_values )
	{
		my $found = FALSE;

		for( my $i=0; $i<scalar( @$ra_list ); ++$i )
		{
			my ( $cmp, $b ) = ( $h_args{ -cmp }, $$ra_list[$i] );
# 			my $ok = eval "\$val $cmp \$b";

			# more simple, i use it until i found a problem
			# works for numbers and strings
			my $ok = ( $val eq $b );
			if( 1 == $ok )
			{
				push( @a_out, $i );
				$found = &TRUE;
				last;
			}
		}

		push( @a_out, undef )	unless( &TRUE == $found );
	}

	return wantarray ? @a_out : $a_out[0];
}


=head2 Procedure Array::Init_OBSOLETE

 Title        : Array::Init_OBSOLETE
 Usage        : &Array::Init( $ra_array, val=>$val, nb_val => $nb );
 Prerequisite : none
 Procedure    : this function initializes an array
 Args         : $ra_array, a ref to the array to modify
               	$val, the value to affect to elements of the array, default is 0
               	$nb, the number of value to initialize _usefull if your array as a size of 0 !
               		default is the length of the array.

 TODO         :	I will wait to need it to implement it.
                	I don't like it that much the way it is

=cut

sub Init_OBSOLETE
{
	my $ra_array = shift;
	my %h_args = ( val => 0,
				nb_val => scalar( @$ra_array ),
				@_ );

	my $key;
	for( my $i=0; $i<$h_args{ nb_val }; ++$i )
	{
		$$ra_array[ $i ] = $h_args{ val };
	}

	return;
}


=head2 Procedure Array::Purge_OBSOLETE

  Title        :	Array::Purge_OBSOLETE
  Usage        :	&Array::Purge( $ra_array );
  Prerequisite :	the values will be temporary duplicated
  Procedure    :	erase undef values of $ra_array.
                	will replace the array ref by a one without undef values
  Args         :	$ra_array, an array ref that might contains undef values

=cut

sub Purge_OBSOLETE
{
	my $ra_array = shift;
	my %h_args = (
			-val => undef,
			@_ );
	my @a_new = ();

	foreach my $value ( @$ra_array )
	{
		push( @a_new, $value )	unless( $h_args{ -val } == $value );
	}

	# destruction de l'ancien tableau et recuperation du nouveau
	@$ra_array = @a_new;

	return;
}

=head2 Function IsDefined_OBSOLETE

	Title      :	IsDefined
	Usage      :	print "\$var n'est pas vide\n"	if( IsDefined( $var ) );
	Prerequiste:	only works with string, any ideas to enhaced with interger (!=0) is welcome
	Function   :	shortcut to know if a variable is defined and non empty
	Returns    :	1 if var is defined and non empty, 0 otherwise
	Args       :	$var, a variable that should contains a string.
	Error      :    none
	Globals    :    none

=cut

sub IsDefined_OBSOLETE
{
	my ($var) = @_;

	return( defined $var and $var ne '');
}
1;
=head2 Function SetUnlessDefined_OBSOLETE

	Title      :	SetUnlessDefined
	Usage      :	SetUnlessDefined( \$var, 'some value' );	# prefer this one
	            	$var = SetUnlessDefined( $var, 'some value' );
	Prerequiste:	none
	Function   :	give a value to a variable if it hasn't got any.
	Returns    :	the value of the variable if a scalar is given as the first arg.
	            	nothing otherwise.
	Args       :	$r_var, a reference variable.
	Error      :    none
	Globals    :    none

=cut

sub SetUnlessDefined_OBSOLETE
{
	my ( $r_var, $val ) = @_;

	if( ref( $r_var ) eq 'SCALAR' )
	{
		$$r_var = $val	unless defined ( $$r_var );
		return;
	}
	elsif( ref( $r_var ) eq '' )
	{
		return  $val	unless IsDefined( $r_var );
		return $r_var;
	}

	return; # cosmetique
}

=head2 Procedure PurgeArray_OBSOLETE

 Title        : PurgeArray_OBSOLETE
 Note         :	Deprecated --> use Array::Purge

=cut

sub PurgeArray_OBSOLETE
{
	warn "obsolete use of General::PurgeArray, prefer Array::Purge\n" . caller();
	my ( $ra_array ) = @_;

	my @a_new = ();

	foreach my $value ( @$ra_array )
	{
		push( @a_new, $value )	if( defined( $value ) );
	}

	# destruction de l'ancien tableau et recuperation du nouveau
	@$ra_array = @a_new;

	return;
}

__END__

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

